//#define LOG_NDEBUG 0
#define LOG_TAG "CameraMonitor"
#include <log/log.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sysutils/NetlinkEvent.h>
#include <hardware_legacy/uevent.h>
#include "CameraMonitor.h"
#include "Debug.h"

namespace android_camera_hal3
{

CameraMonitor* CameraMonitor::sInstance = NULL;

extern "C" {

    using namespace android;
    static void handleUevent(void* data, const char* msg, int len)
    {
        // ALOGV("%s[%d]: data:%p msg:%s len:%d", __FUNCTION__, __LINE__, data, (char*)msg, len);
        CameraMonitor* monitor = (CameraMonitor*)data;
        NetlinkEvent* evt = new NetlinkEvent();

        bool ret = evt->decode((char*)msg, len);
        if(ret == false) {
            ALOGE("failed to decode uevent");
            delete evt;
            return;
        }

        monitor->handleSwitchEvent(evt);
        delete evt;
    }

} //extern "C"

CameraMonitor* CameraMonitor::Instance()
{
    LOG_FUNCTION_NAME();
    if(!sInstance) {
        sInstance = new CameraMonitor();
    }
    return sInstance;
}

CameraMonitor::CameraMonitor()
    : mCallback(NULL)
{
    LOG_FUNCTION_NAME();
    onInitialize();
}

CameraMonitor::~CameraMonitor()
{
    LOG_FUNCTION_NAME();
    uevent_remove_native_handler(&handleUevent);
}

void CameraMonitor::onInitialize()
{
    LOG_FUNCTION_NAME();
    // use uevent library
    if(uevent_get_fd() <= 0) {
        uevent_init();
    }
    uevent_add_native_handler(&handleUevent, this);
    // start pthread
    pthread_create(&mUeventThread, NULL, pthread_entry, (void*)this);
}

void CameraMonitor::handleSwitchEvent(NetlinkEvent* evt)
{
    // ALOGV("%s[%d]: action:%d subsys:%s", __FUNCTION__, __LINE__, evt->getAction(), evt->getSubsystem());
    // evt->dump();
    const char* subsys = evt->getSubsystem();

    if (subsys != NULL && strcmp(subsys, "video4linux") == 0) {
        ALOGD("%s[%d]: [CAMUPDATE] video4linux update", __FUNCTION__, __LINE__);
        // update camera available
        if(mCallback) {
            mCallback();
        }
    }
}

void* CameraMonitor::pthread_entry(void* __unused ptr)
{
    LOG_FUNCTION_NAME();
    char buf[1024];

    while (true) {
        memset(buf, 0x0, 1024);
        uevent_next_event(buf, 1023);
        // ALOGV("%s[%d]: uevent message: %s", __FUNCTION__, __LINE__, buf);
    }
    return NULL;
}

};
