// #define LOG_NDEBUG 0
#define LOG_TAG     "FormatConvertHelper"

#include <log/log.h>
#include <ui/GraphicBufferMapper.h>
#include <libyuv.h>
#ifdef HAL3_HWC_BUFFER
#include "RtkIpc/SeOp.h"
#include "ExifHelper.h"
#else
#include <stdlib.h>

#include <sys/stat.h>
extern "C" {
#include <jpeglib.h>
#include "jhead.h"
}

#include <math.h>
#endif
#include "FormatConvertHelper.h"
#include "misc.h"
#include "MJPGFmt.h"
#include <ui/Rect.h>
#include "Debug.h"

#include "SwConverter.h"

#define BENCHMARK_SE_SCALE              (0)
#define DUMP_JPEG_META                  (0)
#define DUMP_JPEG_RST                   (0)

#if (SUPPORT_EXIF)
#   define EXIF_RESERVED_BYTES          (64*1024)   // 64kB for exif section
#else
#   define EXIF_RESERVED_BYTES          (0*1024)   // 64kB for exif section
#endif

#define THUMB_BUFFER_SIZE               (500*500*16)
#define THUMB_QUALITY                   (10)

#define IS_YUVBUFFER(x)                 (x==ANDROID_SCALER_AVAILABLE_FORMATS_IMPLEMENTATION_DEFINED || \
                                         x==HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP || \
                                         x==HAL_PIXEL_FORMAT_RTK_W16_H16_YCrCb_420_SP || \
                                         x==HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP || \
                                         x==HAL_PIXEL_FORMAT_YCbCr_420_888)
#ifdef HAL3_HWC_BUFFER
#define IS_RTKYUVBUFFER(x)              (x==HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP || \
                                         x==HAL_PIXEL_FORMAT_RTK_W16_H16_YCrCb_420_SP || \
                                         x==HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP)
#else
#define IS_RTKYUVBUFFER(x)              (x==HAL_PIXEL_FORMAT_YCRCB_420_SP )
#endif

int32_t compressYUVtoJPEG(
    uint8_t* input,
    const int width,
    const int height,
    uint8_t* outbuffer,
    int32_t outBufferSize,
    int32_t quality);

// JPEG encoder declare
int compressYuvImageToJpeg(
    int __unused pixelFormat,
    int width,
    int height,
    void* src,
    void* dst,
    int size,
    int quality)
{
    int outlen = compressYUVtoJPEG(
                     (uint8_t*)src,
                     width,
                     height,
                     (uint8_t*)dst,
                     size,
                     quality);

    return outlen;
}

// using namespace android_camera_hal3;

namespace androidcamera3
{

FormatConvertHelper::FormatConvertHelper(
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper* pGrallocWrapper
#endif
)
    :
#ifdef HAL3_HWC_BUFFER
    mHalPixelFormat(HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP),
#else
    mHalPixelFormat(HAL_PIXEL_FORMAT_YCrCb_420_SP),
#endif
    mJpegMetaDstRawCount(0),
    mJpegMetaSrcRawCount(0),
    mJpegRstCount(0)
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    mGrallocWrapper = pGrallocWrapper;
    mSeHandle = SeInit();
    se_init(&mBufAlignRemove_pAllocDev);
#endif

    mThumbBufSize = THUMB_BUFFER_SIZE;
    mThumbBuffer = (uint8_t*) malloc(mThumbBufSize);
    // mCameraMonitor = CameraMonitor::Instance();
}

FormatConvertHelper::~FormatConvertHelper()
{
    LOG_FUNCTION_NAME();
    if(mBufAlignRemove_pAllocDev) {
        se_uninit(&mBufAlignRemove_pAllocDev);
        mBufAlignRemove_pAllocDev = NULL;
        ALOGV("%s[%d]: release mBufAlignRemove_pAllocDev done", __FUNCTION__, __LINE__);
    }
#ifdef HAL3_HWC_BUFFER
    if(mSeHandle != 0) {
        SeUninit(mSeHandle);
        mSeHandle = 0;
        ALOGV("%s[%d]: release mSeHandle done", __FUNCTION__, __LINE__);
    }
#endif
    if(mThumbBuffer != NULL) {
        free(mThumbBuffer);
        mThumbBuffer = NULL;
    }
}

status_t FormatConvertHelper::convertBuffers(Request* r, BUF_CONVERT_INFO* ext)
{
    LOG_FUNCTION_NAME();
    BUF_CONVERT_INFO srcBufInfo = *ext;
    BUF_CONVERT_INFO dstBufInfo;

    // loop all buffers
    int size = r->sensorBuffers->size();
    for (int i = 0; i < size; i++) {
        memset(&dstBufInfo, 0x0, sizeof(BUF_CONVERT_INFO));
        void* vAddr = (*(r->sensorBuffers))[i].y;
        if(vAddr == NULL) {
            getBufferVAddr(
                (*(r->sensorBuffers))[i].buffer,
                (*(r->sensorBuffers))[i].format,
                (*(r->sensorBuffers))[i].width,
                (*(r->sensorBuffers))[i].height,
                &vAddr);
        }

        if(vAddr != srcBufInfo.vAddr) {
            dstBufInfo.w = (*(r->sensorBuffers))[i].width;
            dstBufInfo.h = (*(r->sensorBuffers))[i].height;
            dstBufInfo.format = (*(r->sensorBuffers))[i].format;
            dstBufInfo.pHandle = (*(r->sensorBuffers))[i].buffer;
            dstBufInfo.vAddr = vAddr; //(*(r->sensorBuffers))[i].y;
            dstBufInfo.phyAddr = (*(r->sensorBuffers))[i].phyAddr;

            if(dstBufInfo.phyAddr == 0) {
                getBufferPhyAddr(
                    dstBufInfo.pHandle,
                    &(dstBufInfo.phyAddr));
            }

            dstBufInfo.bufferSize = (*(r->sensorBuffers))[i].buf_size;
#ifdef HAL3_HWC_BUFFER
            dstBufInfo.allocator = NULL;
#endif
            convertOneBuffer(
                &srcBufInfo,
                &dstBufInfo,
                &(r->mMiscInfo));
        }
    }
    return OK;
}

#ifdef HAL3_HWC_BUFFER
#else
status_t FormatConvertHelper::convertBuffers( Request* r, struct vdIn* videoIn_F)
{
    LOG_FUNCTION_NAME();
    BUF_CONVERT_INFO srcBufInfo;
    BUF_CONVERT_INFO dstBufInfo;

    int index = videoIn_F->buf.index;
    srcBufInfo.w = videoIn_F->width;
    srcBufInfo.h = videoIn_F->height;
    srcBufInfo.format = videoIn_F->formatIn;
    srcBufInfo.vAddr =  videoIn_F->mem[index];

    // loop all buffers
    int size = r->sensorBuffers->size();
    r->dump(__FUNCTION__, __LINE__);
    for (int i = 0; i < size; i++) {
        if((*(r->sensorBuffers))[i].format != HAL_PIXEL_FORMAT_BLOB) { // XCheck:
            int srcW = videoIn_F->width;
            int srcH = videoIn_F->height;
            int dstW = (*(r->sensorBuffers))[i].width;
            int dstH = (*(r->sensorBuffers))[i].height;
            ALOGV("%s[%d]: [%i]: index:%i,  src:[%dx%d]/dst:[%dx%d], format:%d, src: %p, dst: %p", __FUNCTION__, __LINE__,
                  i, videoIn_F->buf.index, srcW, srcH, dstW, dstH, (*(r->sensorBuffers))[i].format,
                  (unsigned char*)videoIn_F->mem[index], (unsigned char*)(*(r->sensorBuffers))[i].y);

            void* dstY = NULL;
            void* dstCb = NULL;
            void* dstCr = NULL;
            getBufferCbCr(((*(r->sensorBuffers))[i].buffer), dstW, dstH, &dstY, &dstCb, &dstCr);
            CHECK(dstCb != NULL && dstCr != NULL && dstY != NULL);
            ALOGV("%s[%d]: y:%p, cb:%p, cr:%p", __FUNCTION__, __LINE__, dstY, dstCb, dstCr);

            // XCheck: Merlin3 use I420ToNV12 (false), Mac5p use I420ToNV21 (ture)
            if ((srcW == dstW) && (srcH == dstH)) {
                if ((*(r->sensorBuffers))[i].format == HAL_PIXEL_FORMAT_YCBCR_420_888) {
#if defined(RTD_285X)
                    yuyv422_to_yuv420sp_libyuv(true, (unsigned char*)videoIn_F->mem[index], (unsigned char*)(*(r->sensorBuffers))[i].y, srcW, srcH);
#else
                    yuyv422_to_yuv420sp_libyuv(false, (unsigned char*)videoIn_F->mem[index], (unsigned char*)(*(r->sensorBuffers))[i].y, srcW, srcH);
#endif
                } else {
                    yuyv422_to_yuv420sp_libyuv(true, (unsigned char*)videoIn_F->mem[index], (unsigned char*)(*(r->sensorBuffers))[i].y, srcW, srcH);
                }
            } else {
                if ((*(r->sensorBuffers))[i].format == HAL_PIXEL_FORMAT_YCBCR_420_888) {
#if defined(RTD_285X)
                    yuyv422_to_yuv420sp_libyuv_scale(true, (unsigned char*)videoIn_F->mem[index], srcW, srcH, (unsigned char*)(*(r->sensorBuffers))[i].y, dstW, dstH);
#else
                    yuyv422_to_yuv420sp_libyuv_scale(false, (unsigned char*)videoIn_F->mem[index], srcW, srcH, (unsigned char*)(*(r->sensorBuffers))[i].y, dstW, dstH);
#endif
                } else {
                    yuyv422_to_yuv420sp_libyuv_scale(true, (unsigned char*)videoIn_F->mem[index], srcW, srcH, (unsigned char*)(*(r->sensorBuffers))[i].y, dstW, dstH);
                }
            }
        } else if ((*(r->sensorBuffers))[i].format == HAL_PIXEL_FORMAT_BLOB) {
            ALOGV("%s[%d]: srcBufInfo: [%dx%d], format %d,vAddr %p,videoIn_F->buf.index: %d, videoIn_F:[%dx%d]", __FUNCTION__, __LINE__,
                  srcBufInfo.w, srcBufInfo.h, srcBufInfo.format, srcBufInfo.vAddr, videoIn_F->buf.index, videoIn_F->width, videoIn_F->height);

            memset(&dstBufInfo, 0x0, sizeof(BUF_CONVERT_INFO));
            dstBufInfo.w = (*(r->sensorBuffers))[i].width;
            dstBufInfo.h = (*(r->sensorBuffers))[i].height;
            dstBufInfo.format = (*(r->sensorBuffers))[i].format;
            dstBufInfo.pHandle = (*(r->sensorBuffers))[i].buffer;
            dstBufInfo.phyAddr = (*(r->sensorBuffers))[i].phyAddr;
            dstBufInfo.bufferSize = (*(r->sensorBuffers))[i].buf_size;

            const char* tmpJpgPath = "/tmp/tmp.jpg";
            int fileSize = 0;
            int mJpegQ = 80; // XCheck: Normally, we should use dynamical quality.

            ALOGV("%s[%d]: dstBufInfo: [%dx%d], fmt:%i, size: %i", __FUNCTION__, __LINE__, dstBufInfo.w, dstBufInfo.h, dstBufInfo.format, dstBufInfo.bufferSize);
            unsigned char* pTmpScalYuvBuffer = (unsigned char*)malloc(dstBufInfo.w * dstBufInfo.h * 2); // XChecked: Here need check!!!
            if (pTmpScalYuvBuffer != NULL) {
                yuyv422_scale_libyuv((unsigned char*)videoIn_F->mem[index], videoIn_F->width, videoIn_F->height, (unsigned char*)pTmpScalYuvBuffer, dstBufInfo.w, dstBufInfo.h);
                fileSize = savePicture((unsigned char*)pTmpScalYuvBuffer, dstBufInfo.w, dstBufInfo.h, tmpJpgPath, mJpegQ);
                free(pTmpScalYuvBuffer);
            }

            MISC_FIELD_INFO* info = &(r->mMiscInfo);
            const char* tmpJpgThumb = "/tmp/tmp_thumb.jpg";
            unsigned char* pThumbYuvBuffer = NULL;
            if (info->thumbWidth > 0 && info->thumbHeight > 0) {
                pThumbYuvBuffer = (unsigned char*)malloc(info->thumbWidth * info->thumbHeight * 2); // XChecked: Here need check!!!
            } else {
                ALOGV("%s[%d]: thumb: [%dx%d]", __FUNCTION__, __LINE__, info->thumbWidth, info->thumbHeight);
            }
            if (pThumbYuvBuffer != NULL) {
                int mThumbQ = 10; // XChecked: Here use 10 for minimize JPEG size. Fixed 80 is ok. Normally, we should use dynamical quality.
                ALOGV("%s[%d]: before saveThumbnail info->thumbWidth %d,info->thumbHeight %d", __FUNCTION__, __LINE__, info->thumbWidth, info->thumbHeight);
                yuyv422_scale_libyuv((unsigned char*)videoIn_F->mem[index], videoIn_F->width, videoIn_F->height, (unsigned char*)pThumbYuvBuffer, info->thumbWidth, info->thumbHeight);
                if (savePicture((unsigned char*)pThumbYuvBuffer, info->thumbWidth, info->thumbHeight, tmpJpgThumb, mThumbQ) > 0) {
                    ALOGV("%s[%d]: saveThumbnail done", __FUNCTION__, __LINE__);
                    Addexif(tmpJpgPath, tmpJpgThumb, dstBufInfo.w, dstBufInfo.h, &(r->mMiscInfo));
                    appendThumb(tmpJpgPath, tmpJpgThumb);
                }
                // commitChanges(tmpJpgPath);
                free(pThumbYuvBuffer);
            }

            if (fileSize > 0) {
                ALOGV("fileSize: %i", fileSize);
                FILE* input;
                input = fopen(tmpJpgPath, "rb");
                if (input == NULL) {
                    ALOGE("%s[%d]: savePicture error: Input file == NULL", __FUNCTION__, __LINE__);
                } else {
                    fseek(input, 0L, SEEK_END);
                    fileSize = ftell(input);

                    ALOGV("take picture before lock buffer, fileSize:%i", fileSize);
                    getBufferVAddr(
                        dstBufInfo.pHandle,
                        dstBufInfo.format,
                        fileSize, 1,
                        &dstBufInfo.vAddr);
                    ALOGV("take picture after lock buffer");

                    fseek(input, 0L, SEEK_SET);
                    fread((uint8_t*)dstBufInfo.vAddr, 1, fileSize, input);
                    fclose(input);
                    ALOGV("%s: savePicture success", __FUNCTION__);
                }
            } else {
                ALOGE("%s: savePicture error: fileSize < 0", __FUNCTION__);
            }
        }
    }
    return OK;
}

void FormatConvertHelper::compressToJPEG(BUF_CONVERT_INFO* src, BUF_CONVERT_INFO* dst, MISC_FIELD_INFO* info)
{
    const char* tmpJpgPath = "/tmp/tmp.jpg";
    int fileSize = 0;
    int mJpegQ = 80; // XCheck: Normally, we should use dynamical quality.

    ALOGV("dstBufInfo: [%dx%d], fmt:%i, size: %i", dst->w, dst->h, dst->format, dst->bufferSize);

    unsigned char* pTmpScalYuvBuffer = (unsigned char*)malloc(dst->w * dst->h * 2); // XChecked: Here need check!!!
    unsigned char* pTmpYuvBuffer = (unsigned char*)malloc(src->w * src->h * 2);
    ALOGV("%s[%d]: src:[%ix%i], dst:[%ix%i], srcfmt:%d, dstfmt:%d, srcsize:%d, dstsize:%d", __FUNCTION__, __LINE__, src->w, src->h, dst->w, dst->h, src->format, dst->format, src->bufferSize, dst->bufferSize);
    if ((pTmpScalYuvBuffer != NULL) && (pTmpYuvBuffer != NULL)) {
        yuv420sp_to_yuyv422_libyuv((unsigned char*)src->vAddr, pTmpYuvBuffer, src->w, src->h);
        yuyv422_scale_libyuv(pTmpYuvBuffer, src->w, src->h, (unsigned char*)pTmpScalYuvBuffer, dst->w, dst->h);
        fileSize = savePicture((unsigned char*)pTmpScalYuvBuffer, dst->w, dst->h, tmpJpgPath, mJpegQ);
        free(pTmpScalYuvBuffer);
    }

    const char* tmpJpgThumb = "/tmp/tmp_thumb.jpg";
    unsigned char* pScalThumbYuvBuffer = NULL;
    if (info->thumbWidth > 0 && info->thumbHeight > 0) {
        pScalThumbYuvBuffer = (unsigned char*)malloc(info->thumbWidth * info->thumbHeight * 2);
    } else {
        ALOGV("%s[%d]: thumb: [%dx%d]", __FUNCTION__, __LINE__, info->thumbWidth, info->thumbHeight);
    }
    if ((pScalThumbYuvBuffer != NULL) && (pTmpYuvBuffer != NULL)) {
        int mThumbQ = 10; // XChecked: Here use 10 for minimize JPEG size. Fixed 80 is ok. Normally, we should use dynamical quality.
        ALOGV("before saveThumbnail info->thumbWidth %d,info->thumbHeight %d", info->thumbWidth, info->thumbHeight);
        yuyv422_scale_libyuv(pTmpYuvBuffer, src->w, src->h, (unsigned char*)pScalThumbYuvBuffer, info->thumbWidth, info->thumbHeight);
        if (savePicture((unsigned char*)pScalThumbYuvBuffer, info->thumbWidth, info->thumbHeight, tmpJpgThumb, mThumbQ) > 0) {
            ALOGV("saveThumbnail done ");
            Addexif(tmpJpgPath, tmpJpgThumb, dst->w, dst->h, info);
            appendThumb(tmpJpgPath, tmpJpgThumb);
        }
        // commitChanges(tmpJpgPath);
        free(pScalThumbYuvBuffer);
        free(pTmpYuvBuffer);
    } else if (pTmpYuvBuffer != NULL) {
        free(pTmpYuvBuffer);
    }

    if (fileSize > 0) {
        ALOGV("fileSize: %i", fileSize);
        FILE* input;
        input = fopen(tmpJpgPath, "rb");
        if (input == NULL) {
            ALOGE("%s[%d]: savePicture error: Input file == NULL", __FUNCTION__, __LINE__);
        } else {
            fseek(input, 0L, SEEK_END);
            fileSize = ftell(input);

            getBufferVAddr(dst->pHandle, dst->format, fileSize, 1, &dst->vAddr);
            ALOGV("take picture after lock buffer, fileSize:%i, dst->vAddr %p", fileSize, dst->vAddr);

            fseek(input, 0L, SEEK_SET);
            fread((uint8_t*)dst->vAddr, 1, fileSize, input);
            fclose(input);
            ALOGV("%s: savePicture success", __FUNCTION__);
        }
    } else {
        ALOGE("%s: savePicture error: fileSize < 0", __FUNCTION__);
    }
}

static void saveJPGFile(const char* filename)
{
    char backupName[400];
    struct stat buf;

    memset(backupName, 0, sizeof(char) * 400);
    strncpy(backupName, filename, 395);
    strncat(backupName, ".t", sizeof(backupName) - strlen(backupName) - 1);

    // Remove any .old file name that may pre-exist
    unlink(backupName);

    // Rename the old file.
    rename(filename, backupName);

    // Write the new file.
    WriteJpegFile(filename);
    // Copy the access rights from original file
    if (stat(backupName, &buf) == 0) {
        // set Unix access rights and time to new file
        struct utimbuf mtime;
        chmod(filename, buf.st_mode);

        mtime.actime = buf.st_mtime;
        mtime.modtime = buf.st_mtime;

        utime(filename, &mtime);
    }

    // Now that we are done, remove original file.
    unlink(backupName);
}

void FormatConvertHelper::commitChanges(const char* filename)
{
    if (filename) {
        saveJPGFile(filename);
        DiscardData();
    }
}
int FormatConvertHelper::saveThumbnail(unsigned char* inputBuffer, unsigned char* outputBuffer, int thumbWidth, int thumbHeight, int thumbQ)
{
    if(thumbWidth <= 0 || thumbHeight <= 0) {
        ALOGV("[%d][%s]: mThumb size: [%ix%i], just return", __LINE__, __FUNCTION__, thumbWidth, thumbHeight);
        return 0;
    }
    int thumbW = thumbWidth;
    int thumbH = thumbHeight;
    ALOGV("saveThumbnail before  yuyv422_scale  videoIn_F size: %ix%i outputBuffer: %p", videoIn_F->width, videoIn_F->height, outputBuffer);
    yuyv422_scale_libyuv(inputBuffer, videoIn_F->width, videoIn_F->height, outputBuffer, thumbW, thumbH);
    FILE* output;
    int fileSize;
    int ret;
    const char* filename = "/tmp/tmp_thumb.jpg";
    output = fopen(filename, "wb");
    chmod(filename, S_IRUSR | S_IWUSR  | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

    if (NULL == output) {
        ALOGE("[%d][%s]: Ouput file == NULL", __LINE__, __FUNCTION__);
        return 0;
    }
    // Fix CTS issue: android.hardware.camera2.cts.StillCaptureTest#testJpegExif
    // Because Jhead ThumbLen + ImageInfo.ThumbnailOffset need <= 0x10000-20
    // We need check Jhead and Exif spec or using other exif lib.
    if(thumbQ > 85) {
        thumbQ = 85;
    }

    fileSize = saveYUYVtoJPEG(outputBuffer, thumbW, thumbH, output, thumbQ);

    fclose(output);
    return fileSize;
}

void FormatConvertHelper::insert_thumb (ExifData* ed, ExifLog* log, const char* thumbnail)
{
    FILE* f;
    ALOGV("ed %p", ed);
    if (!ed) {
        return;
    }

    /* Insert new thumbnail */
    f = fopen (thumbnail, "rb");
    if (!f) {
        exif_log (log, EXIF_LOG_CODE_DEBUG, "exif", ("Could not open "
                  "'%s' (%s)!"), thumbnail, strerror (errno));
    } else {
        fseek (f, 0, SEEK_END);
        ed->size = ftell (f);
        ed->data = (unsigned char*)malloc (sizeof (char) * ed->size);
        if (ed->size && !ed->data) {
            EXIF_LOG_NO_MEMORY (log, "exif", sizeof (char) * ed->size);
            exit (1);
        }
        fseek (f, 0, SEEK_SET);
        if (fread (ed->data, sizeof (char), ed->size, f) != ed->size) {
            exif_log (log, EXIF_LOG_CODE_DEBUG, "exif", ("Could not read "
                      "'%s' (%s)."), thumbnail, strerror (errno));
        }
        fclose (f);
    }
}

void FormatConvertHelper::saveJpeg (ExifData* ed, ExifLog* log, const char* filename)
{
    JPEGData* jdata;
    unsigned char* d = NULL;
    unsigned int ds;
    /* Parse the JPEG file. */
    jdata = jpeg_data_new ();
    jpeg_data_log (jdata, log);
    jpeg_data_load_file (jdata, filename);

    /* Make sure the EXIF data is not too big. */
    exif_data_save_data (ed, &d, &ds);
    if (ds) {
        free (d);
        if (ds > 0xffff)
            exif_log (log, EXIF_LOG_CODE_DEBUG, "exif", ("Too much EXIF data "
                      "(%i bytes). Only %i bytes are allowed."),
                      ds, 0xffff);
    };

    jpeg_data_set_exif_data (jdata, ed);

    /* Save the modified image. */
    jpeg_data_save_file (jdata, filename);
    jpeg_data_unref (jdata);

    fprintf (stdout, "Wrote file '%s'.", filename);
    fprintf (stdout, "\n");
}

int FormatConvertHelper::appendThumb(const char* filename, const char* thumbFilename)
{
    int rc = 1;
    ExifLoader* l;
    ExifLog* log = NULL;
    log = exif_log_new ();
    ALOGV("In appendThumb");
    /* Create an ExifLoader object to manage the EXIF loading process */
    l = exif_loader_new();
    exif_loader_log (l, log);
    if (l) {
        ExifData* ed;
        ALOGV("before exif_loader_write_file");
        /* Load the EXIF data from the image file */
        exif_loader_write_file(l, filename);
        ALOGV("before exif_loader_get_data");
        /* Get a pointer to the EXIF data */
        ed = exif_loader_get_data(l);
        ALOGV("before exif_loader_unref");
        /* The loader is no longer needed--free it */
        exif_loader_unref(l);
        l = NULL;

        insert_thumb (ed, log, thumbFilename);
        saveJpeg (ed, log, filename);
        exif_data_unref (ed);
    }
    return rc;
}


/* Create a brand-new tag with a data field of the given length, in the
 * given IFD. This is needed when exif_entry_initialize() isn't able to create
 * this type of tag itself, or the default data length it creates isn't the
 * correct length.
 */
static ExifEntry* create_tag(ExifData* exif, ExifIfd ifd, ExifTag tag, size_t len)
{
    unsigned char* buf;
    ExifEntry* entry;

    /* Create a memory allocator to manage this ExifEntry */
    ExifMem* mem = exif_mem_new_default();
    assert(mem != NULL); /* catch an out of memory condition */

    /* Create a new ExifEntry using our allocator */
    entry = exif_entry_new_mem (mem);
    assert(entry != NULL);

    /* Allocate memory to use for holding the tag data */
    buf = (unsigned char*)exif_mem_alloc(mem, len);
    assert(buf != NULL);

    /* Fill in the entry */
    entry->data = buf;
    entry->size = len;
    entry->tag = tag;
    entry->components = len;
    entry->format = EXIF_FORMAT_UNDEFINED;

    /* Attach the ExifEntry to an IFD */
    exif_content_add_entry (exif->ifd[ifd], entry);

    /* The ExifMem and ExifEntry are now owned elsewhere */
    exif_mem_unref(mem);
    exif_entry_unref(entry);

    return entry;
}
/* Remove spaces on the right of the string */
static void trim_spaces(char* buf)
{
    char* s = buf - 1;
    for (; *buf; ++buf) {
        if (*buf != ' ') {
            s = buf;
        }
    }
    *++s = 0; /* nul terminate the string on the first of the final spaces */
}
static void show_tag(ExifData* d, ExifIfd ifd, ExifTag tag)
{
    ALOGV("In show_tag,tag 0x%x", tag);
    /* See if this tag exists */
    ExifEntry* entry = exif_content_get_entry(d->ifd[ifd], tag);
    if (entry) {
        char buf[1024];

        /* Get the contents of the tag in human-readable form */
        exif_entry_get_value(entry, buf, sizeof(buf));

        /* Don't bother printing it if it's entirely blank */
        trim_spaces(buf);
        if (*buf) {
            ALOGV("%s: %s\n", exif_tag_get_name_in_ifd(tag, ifd), buf);
        }
    } else {
        ALOGV("In show_tag,entry == NULL");
    }
}

int FormatConvertHelper::Addexif(const char* filename, const char* thumbFilename,
                                 int __unused image_jpg_x, int __unused image_jpg_y, MISC_FIELD_INFO* info)
{
    int rc = 1;
    FILE* f;
    unsigned char* exif_data;
    unsigned int exif_data_len;
    ExifEntry* entry;

    FILE* f2 = fopen(filename, "r");
    if (!f2) {
        fprintf(stderr, "Error creating file %s\n", filename);
        ALOGE("Error creating file %s\n", filename);
        return -1;
    }
    fseek (f2, 0, SEEK_END);
    int FileSize  = ftell (f2);
    int image_data_len =  FileSize - image_data_offset;
    fseek (f2, 0, SEEK_SET);

    char* buf = (char*)calloc(FileSize, sizeof(char));
    if(buf == NULL) {
        printf("calloc fails!\n");
        fclose(f2);
        return -1;
    }

    FILE* ft = fopen(thumbFilename, "r");
    if (!ft) {
        fprintf(stderr, "Error creating file %s\n", thumbFilename);
        ALOGE("Error creating file %s\n", thumbFilename);
        return -1;
    }
    fseek (ft, 0, SEEK_END);
    int fileSize  = ftell (ft);
    fseek (ft, 0, SEEK_SET);
    fclose(ft);


    int i = 0;
    int pos = 0;
    char temp = 0;
    for(i = 0; i < FileSize - 1; i++) {
        temp = fgetc(f2);
        if(EOF == (int)temp) {
            break;
        }
        buf[pos++] = temp;
    }
    buf[pos] = 0;
    fclose(f2);

    ExifData* exif = exif_data_new();
    if (!exif) {
        fprintf(stderr, "Out of memory\n");
        free(buf);
        buf = NULL;
        return 2;
    }

    /* Set the image options */
    exif_data_set_option(exif, EXIF_DATA_OPTION_FOLLOW_SPECIFICATION);
    exif_data_set_data_type(exif, EXIF_DATA_TYPE_COMPRESSED);
    exif_data_set_byte_order(exif, FILE_BYTE_ORDER);

    /* Create the mandatory EXIF fields with default data */
    exif_data_fix(exif);

    const char* pucMaker = "Realtek";
    entry = create_tag(exif, EXIF_IFD_0, EXIF_TAG_MAKE, strlen(pucMaker) + 1);
    entry->format = EXIF_FORMAT_ASCII;
    memcpy(entry->data, pucMaker, strlen(pucMaker));
    entry = create_tag(exif, EXIF_IFD_0, EXIF_TAG_MAKE, strlen(pucMaker) + 1);
    entry->format = EXIF_FORMAT_ASCII;
    memcpy(entry->data, pucMaker, strlen(pucMaker));


    //datetime
    struct tm* ltime;
    time_t rawtime;
    time (&rawtime);
    ltime = localtime(&rawtime);
    if (ltime == NULL) {
        ALOGE("[%d][%s]: Error! localtime\n", __LINE__, __FUNCTION__);
    }

    entry = create_tag(exif, EXIF_IFD_0, EXIF_TAG_DATE_TIME,
                       20 * exif_format_get_size(EXIF_FORMAT_ASCII));
    entry->format = EXIF_FORMAT_ASCII;
    entry->components = 20;
    char* pDataTime = (char*)malloc(21);
    sprintf(pDataTime, "%04i:%02i:%02i %02i:%02i:%02i",
            ltime->tm_year + 1900,
            ltime->tm_mon + 1,
            ltime->tm_mday,
            ltime->tm_hour,
            ltime->tm_min,
            ltime->tm_sec);
    memcpy(entry->data, pDataTime, 20);
    free(pDataTime);

    //thumb
    entry = create_tag(exif, EXIF_IFD_1, EXIF_TAG_JPEG_INTERCHANGE_FORMAT,
                       1 * exif_format_get_size(EXIF_FORMAT_LONG));
    entry->format = EXIF_FORMAT_LONG;
    entry->components = 1;
    exif_set_long(entry->data, FILE_BYTE_ORDER, 0);

    entry = create_tag(exif, EXIF_IFD_1, EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH,
                       1 * exif_format_get_size(EXIF_FORMAT_LONG));
    entry->format = EXIF_FORMAT_LONG;
    entry->components = 1;
    exif_set_long(entry->data, FILE_BYTE_ORDER, fileSize);

    entry = create_tag(exif, EXIF_IFD_1, EXIF_TAG_COMPRESSION,
                       1 * exif_format_get_size(EXIF_FORMAT_SHORT));
    entry->format = EXIF_FORMAT_SHORT;
    entry->components = 1;
    exif_set_short(entry->data, FILE_BYTE_ORDER, 6);

    //focallength
    entry = create_tag(exif, EXIF_IFD_EXIF, EXIF_TAG_FOCAL_LENGTH,
                       1 * exif_format_get_size(EXIF_FORMAT_RATIONAL));
    entry->format = EXIF_FORMAT_RATIONAL;
    entry->components = 1;
    ExifRational a;
    a.numerator = 5;
    a.denominator = 1;
    exif_set_rational(entry->data, FILE_BYTE_ORDER, a);

    //ExifRational a;
    ExifRational b;
    ExifRational c;
    //gps
    //if(pGPSinfo != NULL)
    {
        //ALOGV("In addexif ,latitude:%f, longitude:%f, altitude:%d, timestamp:%ld, method:%s"
        //	,info->gpsCoordinates[0],info->gpsCoordinates[1],info->gpsCoordinates[2],info->gpsTimestamp,(const char *)info->gpsProcessingMethod);
        ALOGV("%s[%d]: In addexif ,latitude:%f, longitude:%f, altitude:%lf, timestamp:%lld", __FUNCTION__, __LINE__,
              info->gpsCoordinates[0], info->gpsCoordinates[1], info->gpsCoordinates[2], info->gpsTimestamp);


        //if((param = _params.get(CameraParameters::KEY_GPS_LATITUDE)) != NULL)
        if(info->gpsCoordinates[0] != 0) {
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_LATITUDE_REF, 2);
            entry->format = EXIF_FORMAT_ASCII;
            if(info->gpsCoordinates[0] > 0) {
                memcpy(entry->data, "N", 2);
            } else {
                memcpy(entry->data, "S", 2);
            }
            float latitude_Degree = abs(info->gpsCoordinates[0]);
            float latitude_Cent = (latitude_Degree - (int)latitude_Degree) * 60;
            float latitude_Second = (latitude_Cent - (int)latitude_Cent) * 60;
            a.numerator = latitude_Degree;
            a.denominator = 1;
            b.numerator = (int)latitude_Cent ;
            b.denominator = 1;
            c.numerator = latitude_Second * 1000000;
            c.denominator = 1000000;
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_LATITUDE, 3 * exif_format_get_size(EXIF_FORMAT_RATIONAL));
            entry->format = EXIF_FORMAT_RATIONAL;
            entry->components = 3;
            exif_set_rational(entry->data, FILE_BYTE_ORDER, a);
            exif_set_rational(entry->data + 8, FILE_BYTE_ORDER, b);
            exif_set_rational(entry->data + 16, FILE_BYTE_ORDER, c);
        }

        //if((param = _params.get(CameraParameters::KEY_GPS_LONGITUDE)) != NULL)
        if(info->gpsCoordinates[1] != 0) {
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_LONGITUDE_REF, 2);
            entry->format = EXIF_FORMAT_ASCII;
            entry->components = 2;
            if(info->gpsCoordinates[1] > 0) {
                memcpy(entry->data, "E", 2);
            } else {
                memcpy(entry->data, "W", 2);
            }
            float longitude_Degree = abs(info->gpsCoordinates[1]);
            float longitude_Cent = (longitude_Degree - (int)longitude_Degree) * 60;
            float longitude_Second = (longitude_Cent - (int)longitude_Cent) * 60;
            a.numerator = longitude_Degree;
            a.denominator = 1;
            b.numerator = (int)longitude_Cent ;
            b.denominator = 1;
            c.numerator = longitude_Second * 1000000;
            c.denominator = 1000000;
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_LONGITUDE, 24);
            entry->format = EXIF_FORMAT_RATIONAL;
            entry->components = 3;
            exif_set_rational(entry->data, FILE_BYTE_ORDER, a);
            exif_set_rational(entry->data + 8, FILE_BYTE_ORDER, b);
            exif_set_rational(entry->data + 16, FILE_BYTE_ORDER, c);
        }

        //if((param = _params.get(CameraParameters::KEY_GPS_ALTITUDE)) != NULL)
        if(info->gpsCoordinates[2] != 0) {
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_ALTITUDE_REF, 2);
            entry->format = EXIF_FORMAT_ASCII;
            entry->components = 2;
            if(info->gpsCoordinates[2] > 0) {
                memcpy(entry->data, "0", 2);
            } else {
                memcpy(entry->data, "1", 2);
            }
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_ALTITUDE,
                               1 * exif_format_get_size(EXIF_FORMAT_RATIONAL));
            entry->format = EXIF_FORMAT_RATIONAL;
            entry->components = 1;
            a.numerator = info->gpsCoordinates[2];
            a.denominator = 1;
            exif_set_rational(entry->data, FILE_BYTE_ORDER, a);
        }

        char strDate[15] = {0};
        int timestamp = info->gpsTimestamp;
        int bufLen = sizeof(strDate);
        struct tm tm = *localtime((time_t*)&timestamp);
        strftime(strDate, bufLen - 1, "%Y:%m:%d", &tm);
        strDate[bufLen - 1] = '\0';

        //if((param = _params.get(CameraParameters::KEY_GPS_TIMESTAMP)) != NULL)
        if(info->gpsTimestamp > 0) {
            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_TIME_STAMP,
                               3 * exif_format_get_size(EXIF_FORMAT_RATIONAL));
            entry->format = EXIF_FORMAT_RATIONAL;
            entry->components = 3;
            int sec = info->gpsTimestamp % 60;
            int min = (info->gpsTimestamp % 3600 - sec) / 60;
            int hour = (info->gpsTimestamp % (3600 * 24) - min ) / 24;
            a.numerator = hour;
            a.denominator = 1;
            b.numerator = min;
            b.denominator = 1;
            c.numerator = sec;
            c.denominator = 1;
            exif_set_rational(entry->data, FILE_BYTE_ORDER, a);
            exif_set_rational(entry->data + 8, FILE_BYTE_ORDER, b);
            exif_set_rational(entry->data + 16, FILE_BYTE_ORDER, c);

            entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_DATE_STAMP,
                               11 * exif_format_get_size(EXIF_FORMAT_ASCII));
            entry->format = EXIF_FORMAT_ASCII;
            entry->components = 11;
            char* pDataTime1 = (char*)malloc(12);
            //sprintf(pDataTime1,"%04i:%02i:%02i",tm_year,tm_mon,tm_mday);
            sprintf(pDataTime1, "%s", strDate);
            memcpy(entry->data, pDataTime1, 11);
            free(pDataTime1);
        }

        //if((param = _params.get(CameraParameters::KEY_GPS_PROCESSING_METHOD)) != NULL)
        {
            const char* mExifGpsProcessingMethod = (const char*)(info->gpsProcessingMethod);
            int length = strlen(mExifGpsProcessingMethod);
            if(length > 0) {
                entry = create_tag(exif, EXIF_IFD_GPS, (ExifTag)EXIF_TAG_GPS_PROCESSING_METHOD,
                                   length * exif_format_get_size(EXIF_FORMAT_UNDEFINED));
                entry->format = EXIF_FORMAT_UNDEFINED;
                entry->components = length;
                memcpy(entry->data, info->gpsProcessingMethod, length);
            }
        }

    }

    show_tag(exif, EXIF_IFD_1, EXIF_TAG_JPEG_INTERCHANGE_FORMAT);
    show_tag(exif, EXIF_IFD_1, EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH);

    /* Get a pointer to the EXIF data block we just created */
    exif_data_save_data(exif, &exif_data, &exif_data_len);
    assert(exif_data != NULL);

    f = fopen(filename, "wb");
    if (!f) {
        fprintf(stderr, "Error creating file %s\n", filename);
        ALOGE("Error creating file %s\n", filename);
        exif_data_unref(exif);
        free(buf);
        buf = NULL;
        return rc;
    }

    /* Write EXIF header */
    if (fwrite(exif_header, exif_header_len, 1, f) != 1) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("1 Error writing to file %s\n", filename);
        goto errout;
    }
    /* Write EXIF block length in big-endian order */
    if (fputc((exif_data_len + 2) >> 8, f) < 0) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("2 Error writing to file %s\n", filename);
        goto errout;
    }
    if (fputc((exif_data_len + 2) & 0xff, f) < 0) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("3 Error writing to file %s\n", filename);
        goto errout;
    }
    /* Write EXIF data block */
    if (fwrite(exif_data, exif_data_len, 1, f) != 1) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("4 Error writing to file %s\n", filename);
        goto errout;
    }
    /* Write JPEG image data, skipping the non-EXIF header */
    if (fwrite(buf + image_data_offset, image_data_len, 1, f) != 1) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("5 Error writing to file %s\n", filename);
        goto errout;
    }
    printf("Wrote file %s\n", filename);
    rc = 0;

    free(buf);
    buf = NULL;
    fflush(f);

errout:
    if (fclose(f)) {
        fprintf(stderr, "Error writing to file %s\n", filename);
        ALOGE("6 Error writing to file %s\n", filename);
        rc = 1;
    }
    /* The allocator we're using for ExifData is the standard one, so use
     * it directly to free this pointer.
     */
    free(exif_data);
    exif_data_unref(exif);

    return rc;
}

int FormatConvertHelper::savePicture(unsigned char* inputBuffer, int width, int height, const char* filename, int jpegQ)
{
    LOG_FUNCTION_NAME();
    if(width <= 0 || height <= 0) {
        ALOGE("%s[%d]: mThumb size: [%ix%i], just return", __FUNCTION__, __LINE__, width, height);
        return 0;
    }
    FILE* output;
    int fileSize;
    int ret;
    output = fopen(filename, "wb");
    chmod(filename, S_IRUSR | S_IWUSR  | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);

    if (NULL == output) {
        ALOGE("[%d][%s]: Ouput file == NULL", __LINE__, __FUNCTION__);
        return 0;
    }
    fileSize = saveYUYVtoJPEG(inputBuffer, width, height, output, jpegQ);
    fclose(output);
    return fileSize;
}

int FormatConvertHelper::saveYUYVtoJPEG (unsigned char* inputBuffer, int width, int height, FILE* file, int jpegQ)
{
    LOG_FUNCTION_NAME();
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    JSAMPROW row_pointer[1];
    unsigned char* line_buffer, *yuyv;
    int z;
    int fileSize;

    line_buffer = (unsigned char*) calloc (width * 3, 1);
    if(!line_buffer) {
        ALOGE("saveYUYVtoJPEG line_buffer == NULL");
        return -1;
    }

    yuyv = inputBuffer;

    cinfo.err = jpeg_std_error (&jerr);
    jpeg_create_compress (&cinfo);
    jpeg_stdio_dest (&cinfo, file);

    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;

    jpeg_set_defaults (&cinfo);
    jpeg_set_quality (&cinfo, jpegQ, TRUE);

    jpeg_start_compress (&cinfo, TRUE);
    z = 0;
    while (cinfo.next_scanline < cinfo.image_height) {
        int x;
        unsigned char* ptr = line_buffer;

        for (x = 0; x < width; x++) {
            int r, g, b;
            int y, u, v;

            if (!z) {
                y = yuyv[0];
            } else {
                y = yuyv[2];
            }

            u = yuyv[1];
            v = yuyv[3];

            r = (1192 * (y - 16) + 1634 * (v - 128) ) >> 10;
            g = (1192 * (y - 16) - 833 * (v - 128) - 400 * (u - 128) ) >> 10;
            b = (1192 * (y - 16) + 2066 * (u - 128) ) >> 10;

            *(ptr++) = (r > 255) ? 255 : ((r < 0) ? 0 : r);
            *(ptr++) = (g > 255) ? 255 : ((g < 0) ? 0 : g);
            *(ptr++) = (b > 255) ? 255 : ((b < 0) ? 0 : b);

            if (z++) {
                z = 0;
                yuyv += 4;
            }
        }

        row_pointer[0] = line_buffer;
        jpeg_write_scanlines (&cinfo, row_pointer, 1);
    }

    jpeg_finish_compress (&cinfo);
    fileSize = ftell(file);
    jpeg_destroy_compress (&cinfo);

    free (line_buffer);

    LOG_FUNCTION_END();
    return fileSize;
}
#endif

status_t FormatConvertHelper::convertBuffers(Request* r)
{
    LOG_FUNCTION_NAME();
    BUF_CONVERT_INFO srcBufInfo;
    BUF_CONVERT_INFO dstBufInfo;

    memset(&srcBufInfo, 0x0, sizeof(BUF_CONVERT_INFO));
    findSrcBufInfo((const Request*)r, &srcBufInfo);

    // loop all buffers
    int size = r->sensorBuffers->size();
    ALOGV("%s[%d]: r->sensorBuffers->size() %d", __FUNCTION__, __LINE__, r->sensorBuffers->size());
    for (int i = 0; i < size; i++) {
        memset(&dstBufInfo, 0x0, sizeof(BUF_CONVERT_INFO));
        void* vAddr = NULL;
        getBufferVAddr((*(r->sensorBuffers))[i].buffer, (*(r->sensorBuffers))[i].format, (*(r->sensorBuffers))[i].width, (*(r->sensorBuffers))[i].height, &vAddr);
        ALOGV("%s[%d]: (*(r->sensorBuffers))[%i]: [%dx%d], fmt %d", __FUNCTION__, __LINE__,
              i, (*(r->sensorBuffers))[i].width, (*(r->sensorBuffers))[i].height, (*(r->sensorBuffers))[i].format);
        if (vAddr != srcBufInfo.vAddr) {
            ALOGV("%s[%d]: vAddr != srcBufInfo.vAddr", __FUNCTION__, __LINE__);
            dstBufInfo.w = (*(r->sensorBuffers))[i].width;
            dstBufInfo.h = (*(r->sensorBuffers))[i].height;
            dstBufInfo.format = (*(r->sensorBuffers))[i].format;
            dstBufInfo.pHandle = (*(r->sensorBuffers))[i].buffer;
            dstBufInfo.vAddr = vAddr; //(*(r->sensorBuffers))[i].y;
            dstBufInfo.phyAddr = (*(r->sensorBuffers))[i].phyAddr;
            if (dstBufInfo.phyAddr == 0) {
                getBufferPhyAddr(dstBufInfo.pHandle, &(dstBufInfo.phyAddr));
            }

            dstBufInfo.bufferSize = (*(r->sensorBuffers))[i].buf_size;
#ifdef HAL3_HWC_BUFFER
            dstBufInfo.allocator = NULL;
#endif
            convertOneBuffer(&srcBufInfo, &dstBufInfo, &(r->mMiscInfo));
        }
    }

    return OK;
}

status_t FormatConvertHelper::convertOneBuffer(BUF_CONVERT_INFO* src, BUF_CONVERT_INFO* dst, MISC_FIELD_INFO* info)
{
    LOG_FUNCTION_NAME();
#if 1
    ALOGV("\t%s format %d", __FUNCTION__, dst->format);
#endif

    switch(dst->format) {
#ifdef HAL3_HWC_BUFFER
        case HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP:
        case HAL_PIXEL_FORMAT_RTK_W16_H16_YCrCb_420_SP:
        case HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP:
#endif
        case HAL_PIXEL_FORMAT_YCRCB_420_SP:
            ALOGV("dst->format HAL_PIXEL_FORMAT_YCRCB_420_SP");
            CHECK(OK == convertToRtkRawBuffer(src, dst));
            break;
        case HAL_PIXEL_FORMAT_YCbCr_420_888:
            ALOGV("dst->format HAL_PIXEL_FORMAT_YCbCr_420_888");
            CHECK(OK == convertToNativeRawBuffer(src, dst));
            break;
        case HAL_PIXEL_FORMAT_BLOB:
            ALOGV("dst->format HAL_PIXEL_FORMAT_BLOB");
            CHECK(OK == convertToJPEG(src, dst, info));
            break;
        default:
            ALOGE("%s unsupported format %d", __FUNCTION__, dst->format);
            CHECK(false);
    }

    return OK;
}

status_t FormatConvertHelper::convertToRtkRawBuffer(BUF_CONVERT_INFO* src, BUF_CONVERT_INFO* dst)
{
    LOG_FUNCTION_NAME();
    int srcWidth = ROUNDUP16(src->w);
    int srcHeight = ROUNDUP16(src->h);

    int dstWidth = ROUNDUP16(dst->w);
    int dstHeight = ROUNDUP16(dst->h);

    if(srcWidth == dstWidth && srcHeight == dstHeight) {

        //ALOGV("size src:%d dst:%d",src->bufferSize,dst->bufferSize);
        //CHECK(src->bufferSize <= dst->bufferSize);
#ifdef HAL3_HWC_BUFFER
        SeMemcpy(dst->phyAddr,
                 src->phyAddr,
                 dst->bufferSize, //src->bufferSize,
                 mSeHandle);
#endif
        return OK;
    } else {
        //CHECK((srcWidth*srcHeight) > (dstWidth*dstHeight));

        ALOGV("%s downscale RtkRawBuffer %dx%d -> %dx%d",
              __FUNCTION__,
              src->w,
              src->h,
              dst->w,
              dst->h);

        uint8_t* scaledBuf = NULL;
        uint32_t scaledSize = 0;

        bool rst = ScaleYuv(
                       src->phyAddr,
                       &scaledBuf,
                       &scaledSize,
                       dst->w,
                       dst->h,
                       src->w,
                       src->h);

        if(rst && scaledBuf != NULL && scaledSize > 0) {
            CHECK(dst->bufferSize >= (int32_t)scaledSize);
            memcpy(dst->vAddr, scaledBuf, scaledSize);
            CHECK(dst->pHandle != NULL);
#ifdef HAL3_HWC_BUFFER
            mGrallocWrapper->Flush(*(dst->pHandle));
#else
            ALOGV("mGrallocWrapper->Flush");
#endif
        }

        if(scaledBuf != NULL && scaledSize > 0) {
            free(scaledBuf);
        }
        return OK;
    }

    return INVALID_OPERATION;
}

status_t FormatConvertHelper::convertToNativeRawBuffer(BUF_CONVERT_INFO* src, BUF_CONVERT_INFO* dst)
{
    LOG_FUNCTION_NAME();
    int srcWidth = ROUNDUP16(src->w);
    int srcHeight = ROUNDUP16(src->h);

    int dstWidth = ROUNDUP16(dst->w);
    int dstHeight = ROUNDUP16(dst->h);

    uint8_t* srcVAddr = (uint8_t*)src->vAddr;

    uint8_t* scaledBuf = NULL;
    uint32_t scaledSize = 0;
    bool needDownScale = false;

    //CHECK((srcWidth*srcHeight)>=(dstWidth*dstHeight));

#ifdef HAL3_HWC_BUFFER
    if (src->pHandle != NULL) {
        mGrallocWrapper->Invalidate(*(src->pHandle));
    }

    if (src->allocator != NULL) {
        src->allocator->Invalidate();
    }
#else
    ALOGV("%s[%d]: src [%dx%d], , srcfmt:%d, dst[%dx%d], dstfmt:%d", __FUNCTION__, __LINE__, srcWidth, srcHeight, dstWidth, dstHeight, src->format, dst->format);
#endif

    if (srcWidth * srcHeight > dstWidth * dstHeight) {
        needDownScale = true;
    }

    if (needDownScale) {
        bool rst = ScaleYuv(src->phyAddr, &scaledBuf, &scaledSize, dst->w, dst->h, src->w, src->h);
        if(rst && scaledBuf != NULL && scaledSize > 0) {
            // reset srcVAddr
            srcVAddr = scaledBuf;
        } else {
            ALOGE("%s[%d]: scaleYuv error", __FUNCTION__, __LINE__);
            if(scaledBuf != NULL && scaledSize > 0) {
                free(scaledBuf);
            }
            return INVALID_OPERATION;
        }
    }

    void* dstY = NULL;
    void* dstCb = NULL;
    void* dstCr = NULL;

    getBufferCbCr(dst->pHandle, dst->w, dst->h, &dstY, &dstCb, &dstCr);
    CHECK(dstCb != NULL && dstCr != NULL && dstY != NULL);
    ALOGV("%s[%d]: fmt:%d, y:%p, cb:%p, cr:%p", __FUNCTION__, __LINE__, dst->format, dstY, dstCb, dstCr);

    // support I420 only now
    uint8_t* srcY = srcVAddr;
    uint8_t* srcUV = srcVAddr + ((ROUNDUP16(dst->w)) * (ROUNDUP16(dst->h)));

    // src buffer is 16 aligned
    int srcYStride = ROUNDUP16(dst->w);
    int srcUVStride = ROUNDUP16(dst->w);

    int dstYStride = dst->w;
    int dstUVStride = dst->w / 2;

    libyuv::NV12ToI420(srcY, srcYStride, srcUV, srcUVStride, (uint8_t*)dst->vAddr, dstYStride, (uint8_t*)dstCb, dstUVStride, (uint8_t*)dstCr, dstUVStride, dst->w, dst->h);

    // flush dst buffer
#ifdef HAL3_HWC_BUFFER
    mGrallocWrapper->Flush(*(dst->pHandle));
#endif
    // finally, free scaled buffer
    if(scaledBuf != NULL && scaledSize > 0) {
        free(scaledBuf);
    }
    return OK;
}

status_t FormatConvertHelper::processThumbnail(
    BUF_CONVERT_INFO* src,
    MISC_FIELD_INFO* info,
    uint8_t* pThumbBuf,
    int32_t bufSize,
    int32_t* pByteUsed)
{
    LOG_FUNCTION_NAME();
    int32_t w = info->thumbWidth;
    int32_t h = info->thumbHeight;

    CHECK(w * h > 0);
    *pByteUsed = 0;

    uint8_t* scaled_yuv = NULL;
    uint32_t scaled_buf_size = 0;

    bool rst = ScaleYuvDstNoAlign(
                   (uint32_t)src->phyAddr,
                   &scaled_yuv,
                   &scaled_buf_size,
                   w,
                   h,
                   src->w,
                   src->h);

    if(scaled_yuv != NULL && scaled_buf_size > 0) {

        if(rst) {
            int jpg_buf_size = bufSize;
            int jpegSize = compressYuvImageToJpeg(
                               mHalPixelFormat,        // format
                               w,                      // width
                               h,                      // height
                               scaled_yuv,             // src buffer : se allocated buffer
                               pThumbBuf,              // dst buffer
                               jpg_buf_size,           // dst buffer size
                               THUMB_QUALITY);         // compress quality

            CHECK(jpegSize <= bufSize);
            *pByteUsed = jpegSize;
        }

        free(scaled_yuv);
    }
    return OK;
}

status_t FormatConvertHelper::convertToJPEG(
    BUF_CONVERT_INFO* src,
    BUF_CONVERT_INFO* dst,
    MISC_FIELD_INFO* info)
{
    LOG_FUNCTION_NAME();
    uint8_t* scaled_yuv = NULL;
    uint32_t scaled_buf_size = 0;

    bool rst = ScaleYuvDstNoAlign(
                   (uint32_t)src->phyAddr,
                   &scaled_yuv,
                   &scaled_buf_size,
                   dst->w,
                   dst->h,
                   src->w,
                   src->h);

    if(scaled_yuv != NULL && scaled_buf_size > 0) {

#if (DUMP_JPEG_META)
        // dump dst raw buffer
        if(mJpegMetaDstRawCount < 50) {
            char path[64] = {'\0'};
            sprintf(path, "/tmp/jpg-meta-dst-%d-%dx%d.raw",
                    mJpegMetaDstRawCount,
                    dst->w,
                    dst->h);

            ALOGV("save %s", path);
            dumpData(scaled_yuv,
                     scaled_buf_size,
                     path,
                     &mJpegMetaDstRawCount);
        }

        // dump src raw buffer
        if(mJpegMetaSrcRawCount < 50) {
            char path[64] = {'\0'};
            sprintf(path, "/tmp/jpg-meta-src-%d-%dx%d.raw",
                    mJpegMetaSrcRawCount,
                    src->w,
                    src->h);

            ALOGV("save %s", path);
            dumpData(src->vAddr,
                     src->bufferSize,
                     path,
                     &mJpegMetaSrcRawCount);
        }
#endif

        if(rst) {
            int exif_data_area = (info->thumbWidth * info->thumbHeight > 0) ? EXIF_RESERVED_BYTES : 0;
            int jpg_buf_size = dst->bufferSize - exif_data_area;

            int jpegSize;
            jpegSize = compressYuvImageToJpeg(
                           mHalPixelFormat,        // format
                           dst->w,                 // width
                           dst->h,                 // height
                           scaled_yuv,             // src buffer : se allocated buffer
                           (uint8_t*)dst->vAddr,   // dst buffer
                           jpg_buf_size,           // dst buffer size
                           100);                   // compress quality

            ALOGV("compressYuvImageToJpeg jpegSize %d (available area %d) dst buf %p",
                  jpegSize,
                  jpg_buf_size,
                  dst->vAddr);

            CHECK(jpegSize <= jpg_buf_size);

#if (DUMP_JPEG_RST)
            if(mJpegRstCount < 50) {
                char path[64] = {'\0'};
                sprintf(path, "/tmp/jpg-rst-%d-%dx%d.jpg",
                        mJpegRstCount,
                        dst->w,
                        dst->h);

                ALOGV("save %s", path);
                dumpData(dst->vAddr,
                         jpegSize,
                         path,
                         &mJpegRstCount);
            }
#endif

            ALOGV("%s Thumbnail %dx%d",
                  __FUNCTION__,
                  info->thumbWidth,
                  info->thumbHeight);

            uint8_t* pThumbBuf = NULL;
            int32_t thumbSize = 0;

            if(info->thumbWidth * info->thumbHeight > 0) {
                pThumbBuf = (uint8_t*) malloc(THUMB_BUFFER_SIZE);
                processThumbnail(
                    src,
                    info,
                    pThumbBuf,
                    THUMB_BUFFER_SIZE,
                    &thumbSize);
                ALOGV("%s thumbSize %d", __FUNCTION__, thumbSize);
            }

            JPEG_RST_INFO jpegRst;
            memset(&jpegRst, 0x0, sizeof(JPEG_RST_INFO));

            jpegRst.vaddr = (uint8_t*) dst->vAddr;
            jpegRst.size = dst->bufferSize;
            jpegRst.jpegSize = jpegSize;
            jpegRst.frameNumber = 0xff; // don't care
            jpegRst.pts = 0xff; // don't care
            jpegRst.mSuccess = true;
            jpegRst.mHasJPEG = true;
            jpegRst.pThumbBuf = pThumbBuf;
            jpegRst.thumbSize = thumbSize;

            // allocate work buffer
            uint8_t* wrkBuf = (uint8_t*) malloc(dst->bufferSize);
#ifdef HAL3_HWC_BUFFER
            // add Exif
            add_exif(&jpegRst, wrkBuf, info);
#endif

            // free allocated buffer
            if(pThumbBuf != NULL) {
                free(pThumbBuf);
            }

            if(wrkBuf != NULL) {
                free(wrkBuf);
            }
        }
        free(scaled_yuv);
    }
    return OK;
}

void FormatConvertHelper::getBufferVAddr(buffer_handle_t* buffer, int format, int w, int h, void** ppVPtr)
{
    LOG_FUNCTION_NAME();
    if (HAL_PIXEL_FORMAT_BLOB == format) {
        ALOGV("%s[%d]: format BLOB: 0x%x", __FUNCTION__, __LINE__, format);
#ifdef HAL3_HWC_BUFFER
        int usage = mGrallocWrapper->GetUsage(*buffer);
#else
        int usage = CAMERA_GRALLOC_USAGE; // XChecked: usage
#endif
        const Rect rect(w, h);
        GraphicBufferMapper::get().lock(*buffer, usage, rect, ppVPtr);
        GraphicBufferMapper::get().unlock(*buffer);
    } else {
        ALOGV("%s[%d]: format YUV: 0x%x", __FUNCTION__, __LINE__, format);
        void* pCb;
        void* pCr;
        getBufferCbCr(buffer, w, h, ppVPtr, &pCb, &pCr);
    }
}

void FormatConvertHelper::getBufferCbCr(buffer_handle_t* buffer, int w, int h, void** ppVPtr, void** ppCb, void** ppCr)
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    int usage = mGrallocWrapper->GetUsage(*buffer);
#else
    int usage = CAMERA_GRALLOC_USAGE; // XChecked: usage
#endif
    const Rect rect(w, h);
    android_ycbcr ycbcr = android_ycbcr();
    memset(&ycbcr, 0x0, sizeof(android_ycbcr));
    status_t rst = GraphicBufferMapper::get().lockYCbCr(*buffer, usage, rect, &ycbcr);
    *ppVPtr = ycbcr.y;
    *ppCb = ycbcr.cb;
    *ppCr = ycbcr.cr;
    ALOGV("%s[%d]: [%dx%d], usage:0x%x, y:%p, cb:%p, cr:%p, ystride:%d, cstride:%d, chroma_step:%d", __FUNCTION__, __LINE__,
          w, h, usage, ycbcr.y, ycbcr.cb, ycbcr.cr, ycbcr.ystride, ycbcr.cstride, ycbcr.chroma_step);

    GraphicBufferMapper::get().unlock(*buffer);
}

void FormatConvertHelper::getBufferPhyAddr(buffer_handle_t* buffer, uint32_t* phyAddr)
{
    LOG_FUNCTION_NAME();
    uint32_t phy_info[2];
    CHECK(buffer != NULL);
#ifdef HAL3_HWC_BUFFER
    GrallocWrapper_GetPhyAddrs(mGrallocWrapper, *buffer, (unsigned int*)phy_info);
#endif
    *phyAddr = phy_info[0];
}

void FormatConvertHelper::findSrcBufInfo(const Request* r, BUF_CONVERT_INFO* pSrcInfo)
{
    LOG_FUNCTION_NAME();
    if (NULL == r->pMarkedBuf) {
#ifdef HAL3_HWC_BUFFER
        CHECK(r->allocator != NULL);
#endif
        ALOGV("%s[%d]: NULL == r->pMarkedBuf", __FUNCTION__, __LINE__);
        pSrcInfo->w             = r->allocatorWidth;
        pSrcInfo->h             = r->allocatorHeight;
        pSrcInfo->format        = r->allocatorFormat;
        pSrcInfo->vAddr         = r->allocatorVAddr;
        pSrcInfo->phyAddr       = (uint32_t)r->allocatorPhyAddr;
        pSrcInfo->bufferSize    = r->allocatorBufferSize;
        pSrcInfo->pHandle       = NULL;
#ifdef HAL3_HWC_BUFFER
        pSrcInfo->allocator     = r->allocator;
#endif
        ALOGV("%s[%d]: allocator:[%dx%d], fmt:%d, VAddr:%p, PhyAddr:%u, BufferSize: %d", __FUNCTION__, __LINE__,
              r->allocatorWidth, r->allocatorHeight, r->allocatorFormat, r->allocatorVAddr, (uint32_t)r->allocatorPhyAddr, r->allocatorBufferSize);
    } else {
#ifdef HAL3_HWC_BUFFER
        CHECK(NULL == r->allocator);
#endif
        ALOGV("%s[%d]: r->pMarkedBuf != NULL", __FUNCTION__, __LINE__);
        pSrcInfo->w             = r->pMarkedBuf->width;
        pSrcInfo->h             = r->pMarkedBuf->height;
        pSrcInfo->format        = r->pMarkedBuf->format;
        pSrcInfo->vAddr         = r->pMarkedBuf->y;
        pSrcInfo->phyAddr       = r->pMarkedBuf->phyAddr;
        pSrcInfo->bufferSize    = r->pMarkedBuf->buf_size;
        pSrcInfo->pHandle       = r->pMarkedBuf->buffer;
#ifdef HAL3_HWC_BUFFER
        pSrcInfo->allocator     = NULL;
#endif
        ALOGV("%s[%d]: r->pMarkedBuf:[%dx%d], fmt:%d", __FUNCTION__, __LINE__, r->pMarkedBuf->width, r->pMarkedBuf->height, r->pMarkedBuf->format);
        if (pSrcInfo->vAddr == NULL) {
            // find vAddr
            void* pY = NULL;
            void* pCb = NULL;
            void* pCr = NULL;

            getBufferCbCr(r->pMarkedBuf->buffer, r->pMarkedBuf->width, r->pMarkedBuf->height, &pY, &pCb, &pCr);
            pSrcInfo->vAddr = pY;
        }
        if(pSrcInfo->phyAddr == 0) {
            getBufferPhyAddr(pSrcInfo->pHandle, &(pSrcInfo->phyAddr));
        }
    }
}

bool FormatConvertHelper::se_init(alloc_device_t** ppAllocDev)
{
    LOG_FUNCTION_NAME();
    status_t err = 0;
    buffer_handle_t handle = NULL;
    err = hw_get_module(GRALLOC_HARDWARE_MODULE_ID, &mGrallocModule);
    if(err != 0) {
        return false;
    }

    err = gralloc_open(mGrallocModule, ppAllocDev);

    if(err != 0) {
        return false;
    } else {
        return true;
    }
}

void FormatConvertHelper::se_uninit(alloc_device_t** pAllocDev)
{
    LOG_FUNCTION_NAME();
    if(!pAllocDev) {
        return ;
    }
    int iRet = gralloc_close(*pAllocDev);
    *pAllocDev = NULL;
}

bool FormatConvertHelper::ScaleYuvDstNoAlign(
    uint32_t __unused srcPhyBuffer,
    uint8_t** ppYuv,
    uint32_t* scaled_buf_size,
    int iScaleWidthTo,
    int iScaleHeightTo,
    int __unused srcWidth,
    int __unused srcHeight)
{
    ALOGV("mBufAlignRemove_pAllocDev %p", mBufAlignRemove_pAllocDev);

    if(!mBufAlignRemove_pAllocDev) {
        return false;
    }

    // should no align
    ALOGV("iScaleWidthTo %d iScaleHeightTo %d", iScaleWidthTo, iScaleHeightTo);

    if(iScaleWidthTo <= 0 || iScaleHeightTo <= 0) {
        return false;
    }

#if (BENCHMARK_SE_SCALE)
    nsecs_t start_time = systemTime();
#endif

    bool bRet = true;
    buffer_handle_t Buf_hBufferHandle = 0;
    uint32_t Buf_hPhyBufferHandle = 0;  // phyAddr info

    int fmt = mHalPixelFormat;
    /*
    if(isCompressedFormat(mFormat)){
        fmt = JPEG_ENCODER_FMT;
    }else{
        fmt = mHalPixelFormat;
    }
    */

    // allocate dst buffer (put scaled buffer data)
    se_Create420spPhyBuffer(
        mBufAlignRemove_pAllocDev,
        /*ROUNDUP16*/(iScaleWidthTo),
        /*ROUNDUP16*/(iScaleHeightTo),
        fmt,
        &Buf_hBufferHandle,
        &Buf_hPhyBufferHandle);

    if(!Buf_hPhyBufferHandle) {
        return false;
    }

    void* virtaddrScale = NULL;
    se_PhyBuf2VirBuf(
        /*ROUNDUP16*/(iScaleWidthTo),
        /*ROUNDUP16*/(iScaleHeightTo),
        Buf_hBufferHandle,
        &virtaddrScale);

    ALOGV("[%s] virtaddrScale %p", __FUNCTION__, virtaddrScale);
#ifdef HAL3_HWC_BUFFER
    int iRet = SeScalingDstNo16Align(
                   (unsigned int)Buf_hPhyBufferHandle,
                   (unsigned int)srcPhyBuffer,
                   /*ROUNDUP16*/(iScaleWidthTo),
                   /*ROUNDUP16*/(iScaleHeightTo),
                   srcWidth,
                   srcHeight,
                   (unsigned int)mSeHandle,
                   0);

    ALOGV("[ScaleYuvDstNoAlign] iRet %.8x", iRet);
#endif

    // fix garbage green lines
#ifdef HAL3_HWC_BUFFER
    mGrallocWrapper->Invalidate(Buf_hBufferHandle);
#endif
    // NOTE: always converting YUV raw buffer
    int iSize = (iScaleWidthTo) * (iScaleHeightTo) * 12 / 8;

#if 0 //(DUMP_SCALE_DST_NO_ALI)
    if(mDumpFileCntScalNoAli < 50) {
        char path[64] = {'\0'};
        sprintf(path,
                "/tmp/i-scale-no-ali-%d-%dx%d.raw",
                mDumpFileCntScalNoAli,
                iScaleWidthTo,
                iScaleHeightTo);
        ALOGV("save %s", path);
        dumpData((void*)virtaddrScale,
                 iSize,
                 path,
                 &mDumpFileCntScalNoAli);
    }
#endif

    ALOGV("[%s] iSize %d", __FUNCTION__, iSize);
    *ppYuv = (uint8_t*)malloc(iSize);  //mRequestMemory(-1, iSize, 1, NULL);
    CHECK(NULL != *ppYuv);
    memset(*ppYuv, 0x0, iSize);

    *scaled_buf_size = iSize;
    memcpy(*ppYuv, (void*)virtaddrScale, iSize);

    if(Buf_hBufferHandle) {
        se_DestroyPhyBuffer(mBufAlignRemove_pAllocDev, Buf_hBufferHandle);
    }

#if (BENCHMARK_SE_SCALE)
    nsecs_t time_used = systemTime() - start_time;
    ALOGV("SeScalingDstNoAlign use %.2f ms", (float)time_used / 1E6f);
#endif

    return bRet;
}

bool FormatConvertHelper::ScaleYuv(
    uint32_t __unused srcPhyBuffer,
    uint8_t** ppYuv,
    uint32_t* scaled_buf_size,
    int iScaleWidthTo,
    int iScaleHeightTo,
    int __unused srcWidth,
    int __unused srcHeight)
{
    ALOGV("mBufAlignRemove_pAllocDev %p", mBufAlignRemove_pAllocDev);

    if(!mBufAlignRemove_pAllocDev) {
        return false;
    }

    ALOGV("[%s] iScaleWidthTo %d iScaleHeightTo %d",
          __FUNCTION__,
          iScaleWidthTo,
          iScaleHeightTo);

    if(iScaleWidthTo <= 0 || iScaleHeightTo <= 0) {
        return false;
    }

#if (BENCHMARK_SE_SCALE)
    nsecs_t start_time = systemTime();
#endif

    bool bRet = true;
    buffer_handle_t Buf_hBufferHandle = 0;
    uint32_t Buf_hPhyBufferHandle = 0;  // phyAddr info

    int fmt = mHalPixelFormat;
    /*
    if(isCompressedFormat(mFormat)){
        fmt = JPEG_ENCODER_FMT;
    }else{
        fmt = mHalPixelFormat;
    }
    */

    // allocate dst buffer (put scaled buffer data)
    se_Create420spPhyBuffer(
        mBufAlignRemove_pAllocDev,
        ROUNDUP16(iScaleWidthTo),
        ROUNDUP16(iScaleHeightTo),
        fmt,
        &Buf_hBufferHandle,
        &Buf_hPhyBufferHandle);

    if(!Buf_hPhyBufferHandle) {
        return false;
    }

    void* virtaddrScale = NULL;
    se_PhyBuf2VirBuf(
        ROUNDUP16(iScaleWidthTo),
        ROUNDUP16(iScaleHeightTo),
        Buf_hBufferHandle,
        &virtaddrScale);

    ALOGV("[%s] virtaddrScale %p", __FUNCTION__, virtaddrScale);
#ifdef HAL3_HWC_BUFFER
    int iRet = SeScaling(
                   (unsigned int)Buf_hPhyBufferHandle,
                   (unsigned int)srcPhyBuffer,
                   /*ROUNDUP16*/(iScaleWidthTo),
                   /*ROUNDUP16*/(iScaleHeightTo),
                   srcWidth,
                   srcHeight,
                   (unsigned int)mSeHandle,
                   0);

    ALOGV("[ScaleYuv] iRet %.8x", iRet);
    // fix garbage green lines
    mGrallocWrapper->Invalidate(Buf_hBufferHandle);
#endif
    // NOTE: always converting YUV raw buffer
    int iAlignSize = ROUNDUP16(iScaleWidthTo) * ROUNDUP16(iScaleHeightTo) * 12 / 8;

    //dump2file((char*)"/tmp/secacle.nv12", (unsigned char*)virtaddrScale, iAlignSize);
    ALOGV("[%s] iAlignSize %d", __FUNCTION__, iAlignSize);
    *ppYuv = (uint8_t*)malloc(iAlignSize);  //mRequestMemory(-1, iAlignSize, 1, NULL);
    CHECK(NULL != *ppYuv);
    memset(*ppYuv, 0x0, iAlignSize);

    *scaled_buf_size = iAlignSize;
    memcpy(*ppYuv, (void*)virtaddrScale, iAlignSize);

    if(Buf_hBufferHandle) {
        se_DestroyPhyBuffer(mBufAlignRemove_pAllocDev, Buf_hBufferHandle);
    }

#if (BENCHMARK_SE_SCALE)
    nsecs_t time_used = systemTime() - start_time;
    ALOGV("SeScaling use %.2f ms", (float)time_used / 1E6f);
#endif

    return bRet;
}

bool FormatConvertHelper::se_Create420spPhyBuffer(
    alloc_device_t* pAllocDev,
    int iWidth,
    int iHeight,
    int __unused format,
    buffer_handle_t* pBufferHandle,
    uint32_t* pPhyBufferHandle)
{
    LOG_FUNCTION_NAME();
    if(!pAllocDev) {
        return false;
    }

    int32_t stride = 0;
    status_t err = 0;

#if 0
    int usage = GRALLOC_USAGE_HW_ION |
                GRALLOC_USAGE_HW_COMPOSER |
                GRALLOC_USAGE_HW_VIDEO_ENCODER |
                GRALLOC_USAGE_SW_READ_OFTEN |
                GRALLOC_USAGE_SW_WRITE_OFTEN;
#endif

#ifdef HAL3_HWC_BUFFER
    int usage = GRALLOC_USAGE_HW_ION
                | GRALLOC_USAGE_HW_COMPOSER
                | GRALLOC_USAGE_HW_VIDEO_ENCODER;
#endif
    int usage = GRALLOC_USAGE_HW_COMPOSER
                | GRALLOC_USAGE_HW_VIDEO_ENCODER;

    const struct hw_module_t* psModule = NULL;
    err = hw_get_module(GRALLOC_HARDWARE_MODULE_ID, &psModule);
#ifdef HAL3_HWC_BUFFER
    mali_gralloc_private_module_t* psGrallocModule =
        (mali_gralloc_private_module_t*) psModule;
#endif
    uint32_t phys_info[2];
    buffer_handle_t buf;
#ifdef HAL3_HWC_BUFFER
    err = pAllocDev->alloc(
              pAllocDev,
              iWidth,
              iHeight,
              HAL_PIXEL_FORMAT_RTK_W16_H16_YCbCr_420_SP,
              ///*36 */ HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP /*mHalPixelFormat*/,
              usage,
              &buf,
              &stride);
#else
    ALOGV("se_Create420spPhyBuffer 1");
    err = pAllocDev->alloc(
              pAllocDev,
              iWidth,
              iHeight,
              HAL_PIXEL_FORMAT_YCrCb_420_SP,
              ///*36 */ HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP /*mHalPixelFormat*/,
              usage,
              &buf,
              &stride);
#endif
    if(err != NO_ERROR) {
        return false;
    }
#ifdef HAL3_HWC_BUFFER
    err = psGrallocModule->GetPhyAddrs(
              (struct private_module_t const*) psGrallocModule,
              buf,
              phys_info);
#else
    ALOGV("se_Create420spPhyBuffer 2");
    err = NO_ERROR;
#endif
    if(err != NO_ERROR) {
        return false;
    }

    // do some test..
    void* virAddr;
    se_PhyBuf2VirBuf(
        iWidth,
        iHeight,
        buf,
        &virAddr);

    *pPhyBufferHandle = phys_info[0]; //physical address
    ALOGV("se_Create420spPhyBuffer phyAddr %.8x virAddr %p", phys_info[0], virAddr);
    // size is unused....
    // uint32_t size = iWidth * iHeight * V4L2FmtToBpp(format) / 8; //phys_info[1] is page aligned
    *pBufferHandle = buf;
    return true;
}

bool FormatConvertHelper::se_PhyBuf2VirBuf(
    int iWidth,
    int iHeight,
    buffer_handle_t pBufferHandle,
    void** pvirtaddr)
{
    LOG_FUNCTION_NAME();
#ifdef HAL3_HWC_BUFFER
    int usage = GRALLOC_USAGE_HW_ION
                | GRALLOC_USAGE_HW_COMPOSER
                | GRALLOC_USAGE_HW_VIDEO_ENCODER
                | GRALLOC_USAGE_SW_READ_OFTEN
                | GRALLOC_USAGE_SW_WRITE_OFTEN;
#else
    int usage = GRALLOC_USAGE_HW_COMPOSER
                | GRALLOC_USAGE_HW_VIDEO_ENCODER
                | GRALLOC_USAGE_SW_READ_OFTEN
                | GRALLOC_USAGE_SW_WRITE_OFTEN;
#endif
    const gralloc_module_t* grmodule =
        (const gralloc_module_t*)mGrallocModule;

    grmodule->lock(
        grmodule,
        pBufferHandle,
        usage,
        0,
        0,
        iWidth,
        iHeight,
        pvirtaddr);

    return true;
}

bool FormatConvertHelper::se_DestroyPhyBuffer(
    alloc_device_t* pAllocDev,
    buffer_handle_t pBufferHandle)
{
    LOG_FUNCTION_NAME();
    if(!pAllocDev) {
        return false;
    }

    if(pBufferHandle == 0) {
        return false;
    }

    pAllocDev->free(pAllocDev, pBufferHandle);
    pBufferHandle = 0;
    return true;
}

bool FormatConvertHelper::isCompressedFormat(int v4l2Fmt)
{
    if(isCompressedFmt(v4l2Fmt) > 0) {
        return true;
    } else {
        return false;
    }
}

void FormatConvertHelper::dumpData(
    void* vaddr,
    int size,
    char* path,
    uint32_t* pCnt)
{
    FILE* fd = fopen(path, "w");
    if(fd != NULL) {
        fwrite(vaddr, 1, size, fd);
        int rst = fclose(fd);
        ALOGV("[%s] path %s size %d rst %d", __FUNCTION__, path, size, rst);
        (*pCnt)++;
    }
}

int32_t FormatConvertHelper::createThumbnailJpeg(
    uint32_t srcPhyAddr,
    uint32_t srcWidth,
    uint32_t srcHeight,
    int32_t thumbWidth,
    int32_t thumbHeight)
{
    uint8_t* scaled_yuv = NULL;
    uint32_t scaled_buf_size = 0;
    int32_t thumbSize = 0;

    // clear thumbnail buffer
    memset(mThumbBuffer, 0x0, mThumbBufSize);

    // make a small yuv frame
    bool rst = ScaleYuvDstNoAlign(srcPhyAddr,
                                  &scaled_yuv,
                                  &scaled_buf_size,
                                  thumbWidth,
                                  thumbHeight,
                                  srcWidth,
                                  srcHeight);

    ALOGV("[%s] rst %d scaled_yuv %p scaled_buf_size %d",
          __FUNCTION__,
          rst,
          scaled_yuv,
          scaled_buf_size);

    char path[64] = {'\0'};
    if(rst && scaled_yuv != NULL && scaled_buf_size > 0) {

        int fmt = mHalPixelFormat;

        ALOGV("[%s] compressYuvImageToJpeg - fmt %.4x thumbW: %d thumbH: %d",
              __FUNCTION__,
              fmt,
              thumbWidth,
              thumbHeight);

        uint8_t* jpg_buf_vaddr = mThumbBuffer;
        uint32_t jpg_buf_size = mThumbBufSize;

        thumbSize = compressYuvImageToJpeg(
                        fmt,
                        thumbWidth,
                        thumbHeight,
                        scaled_yuv,
                        jpg_buf_vaddr,
                        jpg_buf_size,
                        THUMB_QUALITY);
    }

    // always free buffer, is that correct?
    if(scaled_yuv != NULL && scaled_buf_size > 0) {
        ALOGV("[%s] free scale yuv buf", __FUNCTION__);
        free(scaled_yuv);
    }

    return thumbSize;
}

void FormatConvertHelper::compressJPEG(
    JPEG_SRC_BUF_INFO* pSrc,
    JPEG_DST_BUF_INFO* pDst)
{
    LOG_FUNCTION_NAME();
    uint8_t* scaled_yuv = NULL;
    uint32_t scaled_buf_size = 0;
    int dstWidth = pDst->w;
    int dstHeight = pDst->h;

    bool rst = ScaleYuvDstNoAlign((uint32_t)pSrc->phyAddr,
                                  &scaled_yuv,
                                  &scaled_buf_size,
                                  dstWidth,
                                  dstHeight,
                                  pSrc->w,
                                  pSrc->h);

    if(rst && scaled_yuv != NULL && scaled_buf_size > 0) {
        uint8_t* jpg_buf_vaddr = (uint8_t*) pDst->vaddr;

        int thumbSize = 0;
        bool mNeedThumbnail = false;

        if((pDst->tw * pDst->th) > 0) {
            mNeedThumbnail = true;
        }

        int exif_data_area = mNeedThumbnail ? EXIF_RESERVED_BYTES : 0;
        int jpg_buf_size = pDst->size - exif_data_area;

        int jpegSize = 0;

        jpegSize = compressYuvImageToJpeg(
                       mHalPixelFormat,
                       dstWidth,
                       dstHeight,
                       scaled_yuv,
                       jpg_buf_vaddr,
                       jpg_buf_size,
                       10);

        CHECK(jpegSize <= jpg_buf_size);

        if(mNeedThumbnail) {
            // first prepare thumbnail
            thumbSize = createThumbnailJpeg(
                            (uint32_t)pSrc->phyAddr,
                            pSrc->w,
                            pSrc->h,
                            pDst->tw,
                            pDst->th);
        }

        // report result
        pDst->bytesused = jpegSize;
        pDst->thumbVAddr = mThumbBuffer;
        pDst->thumbBufUsed = thumbSize;
    }

    // always free buffer, is that correct?
    if(scaled_yuv != NULL && scaled_buf_size > 0) {
        //ALOGV("[%s] free scale yuv buf",__FUNCTION__);
        free(scaled_yuv);
    }
}

void FormatConvertHelper::setRtkPrivData(Request* r, uint32_t inputFps)
{
    LOG_FUNCTION_NAME();
    int size = r->sensorBuffers->size();
    for (int i = 0; i < size; i++) {
        StreamBuffer* buf = (StreamBuffer*) & ((*(r->sensorBuffers))[i]);
        setRtkPrivData(buf, inputFps);
    }
}

void FormatConvertHelper::setRtkPrivData(StreamBuffer* pStreamBuf, uint32_t __unused inputFps)
{
    LOG_FUNCTION_NAME();
    if(!IS_RTKYUVBUFFER(pStreamBuf->format)) {
        return;
    }

    int delay_mode = 6; // DISPLAY_ORDER
    int delay_depth = 2;

#if 0
    CHECK(mCameraMonitor != NULL);
    uint32_t txFps = 60;
    if(mCameraMonitor->getHDMIPlugged()) {
        HDMITX_STATUS* pTxStatus = mCameraMonitor->getTxStatus();
        uint32_t fps = pTxStatus->mFps;
        if(fps <= 0) {
            ALOGE("No TxFPS info");
        }
        txFps = fps;
    }

    if(txFps >= inputFps) {
        delay_mode = 6;
        delay_depth = 2;
        ALOGV("delay_mode 6 txFps %d inputFps %d", txFps, inputFps);
    } else {
        ALOGV("delay_mode 2 txFps %d inputFps %d", txFps, inputFps);
    }
#endif

    ALOGV("%s for buffer format 0x%.2x", __FUNCTION__, pStreamBuf->format);

    memset(&(pStreamBuf->rtkData), 0x0, sizeof(private_rtk_data));

    // always report progress frame info
    pStreamBuf->rtkData.yuv_mode = CONSECUTIVE_FRAME;
    pStreamBuf->rtkData.delay_mode = delay_mode;
    pStreamBuf->rtkData.delay_depth = delay_depth;
    pStreamBuf->rtkData.init_frame = 1;

    pStreamBuf->rtkData.lumaOffTblAddr = 0xffffffff;
    pStreamBuf->rtkData.chromaOffTblAddr = 0xffffffff;
#ifdef HAL3_HWC_BUFFER
    mali_gralloc_private_module_t* psGrallocModule =
        (mali_gralloc_private_module_t*)mGrallocModule;

    int ret = psGrallocModule->SetPrivData(
                  (struct private_module_t const*) psGrallocModule,
                  *(pStreamBuf->buffer),
                  (void*) & (pStreamBuf->rtkData),
                  sizeof(pStreamBuf->rtkData));
#endif
}

bool FormatConvertHelper::getPrivData(StreamBuffer* pStreamBuf)
{
    LOG_FUNCTION_NAME();
    memset(&(pStreamBuf->rtkDataReturn), 0x0, sizeof(private_rtk_data));
#ifdef HAL3_HWC_BUFFER
    mali_gralloc_private_module_t* psGrallocModule =
        (mali_gralloc_private_module_t*)mGrallocModule;

    int ret = psGrallocModule->GetPrivData(
                  (struct private_module_t const*) psGrallocModule,
                  *(pStreamBuf->buffer),
                  (void*)(&(pStreamBuf->rtkDataReturn)),
                  sizeof(private_rtk_data));
#else
    int ret = 0;
    ALOGV("FormatConvertHelper::getPrivData");
#endif

    if(ret != 0) {
        return false;
    }

    return true;
}

};
