//#define LOG_NDEBUG 0
#define LOG_TAG "misc"

#include <stdlib.h>
#include <string.h>
#include <linux/videodev2.h>
#include "CameraParameters.h"
#include <system/graphics.h>
#include "misc.h"
#include "Debug.h"

namespace android
{

int CameraFmtToV4L2(const char* fmt)
{
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_YUV422I)) {
        return V4L2_PIX_FMT_YUYV;
    }
    //if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_YVU420SP))
    //    return V4L2_PIX_FMT_NV21;
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_YUV420SP)) {
        return V4L2_PIX_FMT_NV21;
    }
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_YUV420P)) {
        return V4L2_PIX_FMT_YVU420;
    }
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_JPEG)) {
        return V4L2_PIX_FMT_MJPEG;
    }
#ifdef HAL3_HWC_BUFFER
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_H264)) {
        return V4L2_PIX_FMT_H264;
    }
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_RGB565)) {
        return V4L2_PIX_FMT_RGB565;
    }
    if(!strcmp(fmt, CameraParameters::PIXEL_FORMAT_RGBA8888)) {
        return V4L2_PIX_FMT_RGB32;
    }
#endif
    ALOGE("CameraFmtToV4L2: invalid format:%c%c%c%c", FCC2ASCII(fmt));
    return 0;
}

const char* V4L2FmtToStr(int fmt)
{
    switch(fmt) {
        case V4L2_PIX_FMT_YUYV:
            return CameraParameters::PIXEL_FORMAT_YUV422I;
        //case V4L2_PIX_FMT_NV21:
        //    return CameraParameters::PIXEL_FORMAT_YVU420SP;
        case V4L2_PIX_FMT_NV21:
            return CameraParameters::PIXEL_FORMAT_YUV420SP;
        case V4L2_PIX_FMT_YVU420:
            return CameraParameters::PIXEL_FORMAT_YUV420P;
        case V4L2_PIX_FMT_MJPEG:
            return CameraParameters::PIXEL_FORMAT_JPEG;
#ifdef HAL3_HWC_BUFFER
        case V4L2_PIX_FMT_H264:
            return CameraParameters::PIXEL_FORMAT_H264;
        case V4L2_PIX_FMT_RGB565:
            return CameraParameters::PIXEL_FORMAT_RGB565;
        //case V4L2_PIX_FMT_RGB32:
        //    return CameraParameters::PIXEL_FORMAT_RGBA8888;
        case V4L2_PIX_FMT_BGR32:
            return CameraParameters::PIXEL_FORMAT_RGBA8888;
#endif
        default:
            ALOGE("V4L2FmtToStr: no suitable value for format:%c%c%c%c", FCC2ASCII(fmt));
            return NULL;
    }
}

int V4L2FmtToBpp(int fmt)
{
    switch(fmt) {
        case V4L2_PIX_FMT_RGB32:
        case V4L2_PIX_FMT_BGR32:
            return 32;
        case V4L2_PIX_FMT_MJPEG:
        case V4L2_PIX_FMT_H264:
            //return 16; //make buffer size = WxHx2
            return 24; // for zte camera
        case V4L2_PIX_FMT_YUYV:
        case V4L2_PIX_FMT_RGB565:
            return 16;
        case V4L2_PIX_FMT_NV12:
        case V4L2_PIX_FMT_NV21:
        case V4L2_PIX_FMT_YUV420:
        case V4L2_PIX_FMT_YVU420:
            return 12;
        default:
            ALOGE("V4L2FmtToBpp: invalid format:%c%c%c%c", FCC2ASCII(fmt));
            return 1;
    }
}

int V4L2ToHalPixel(int fmt)
{
    LOG_FUNCTION_NAME();
    ALOGV("%s[%d]: fmt %s", __FUNCTION__, __LINE__, V4L2FmtToStr(fmt));

    switch(fmt) {
        case V4L2_PIX_FMT_RGB32: //byte 0~3: ARGB
            return HAL_PIXEL_FORMAT_RGBA_8888; //byte 0~3: RGBA, FIXME: NO MATCH!!!
        case V4L2_PIX_FMT_BGR32: //byte 0~3: BGRA
            return HAL_PIXEL_FORMAT_BGRA_8888; //byte 0~3: BGRA
        case V4L2_PIX_FMT_RGB565:
            return HAL_PIXEL_FORMAT_RGB_565;
        case V4L2_PIX_FMT_YUYV:
#ifdef HAL3_HWC_BUFFER
            return HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP; //HAL_PIXEL_FORMAT_YCbCr_422_888; //HAL_PIXEL_FORMAT_YCbCr_422_I;
#else
            return HAL_PIXEL_FORMAT_YCrCb_420_SP;
#endif
        case V4L2_PIX_FMT_YUV420:
#ifdef HAL3_HWC_BUFFER
            return HAL_PIXEL_FORMAT_YCbCr_420_P;
#else
            return HAL_PIXEL_FORMAT_YCrCb_420_SP;
#endif
        case V4L2_PIX_FMT_YVU420:
            return HAL_PIXEL_FORMAT_YV12;
        case V4L2_PIX_FMT_NV21:
            return HAL_PIXEL_FORMAT_YCrCb_420_SP;
        case V4L2_PIX_FMT_NV12:
#ifdef HAL3_HWC_BUFFER
            return HAL_PIXEL_FORMAT_RTK_16_YCbCr_420_SP;
#else
            return HAL_PIXEL_FORMAT_YCrCb_420_SP;
#endif

        case V4L2_PIX_FMT_MJPEG:
            // TODO:
            return -1;
    }
    ALOGE("%s[%d]: invalid format:%c%c%c%c", __FUNCTION__, __LINE__, FCC2ASCII(fmt));
    return 0;
}

int isCompressedFmt(int fmt)
{
    if(fmt == V4L2_PIX_FMT_MJPEG) {
        return 1;
    }
    if(fmt == V4L2_PIX_FMT_H264) {
        return 1;
    }
    return 0;
}

}; // namespace android
